//     Author: Maurizio Loreti, aka MLO or (HAM) I3NOO
//     Work:   University of Padova - Department of Physics
//             Via F. Marzolo, 8 - 35131 PADOVA - Italy
//     Phone:  +39 (049) 827-7216   FAX: +39 (049) 827-7102
//     EMail:  loreti@pd.infn.it
//     WWW:    http://www.pd.infn.it/~loreti/mlo.html
//
// -------------------------------------------------------------------
// This program queries a trusted time server for the current time,
// and compares the returned value with the time on the local host.
// On request, the time on the local host may be reset.
//
// synctime has been inspired from "rdata"; however, is a project
// independently developed.
// -------------------------------------------------------------------
// $Id: main.cxx,v 1.7 2006/11/24 10:22:34 loreti Exp $
// -------------------------------------------------------------------

#include <stdexcept>

extern "C" {
  #include <errno.h>
  #include <getopt.h>
  #include <sys/time.h>
}

#include "synctime.hh"

using std::cerr;
using std::cout;

namespace {
  bool setTime              = false;
  bool quiet                = false;
  timeSocket::socketType st = timeSocket::TCP;
  const char nist[]         = "time-a.timefreq.bldrdoc.gov";
  const char versionID[]    = "synctime v6 - MLO 15-Feb-2004\n";
}

void usage(
  char * progName
) {
  cerr << "Usage: " << progName << " [options] [host]\n";
  cerr << "  Queries the time port of the given host for the "
       << "current date and time\n\n";
  cerr << "Options:\n";
  cerr << "  -s, --set     : changes the current time on the local "
       << "host\n";
  cerr << "  -t, --tcp     : connects to the remote host via a TCP "
       << "socket (default)\n";
  cerr << "  -u, --udp     : connects to the remote host via an UDP "
       << "socket\n";
  cerr << "  -q, --quiet   : suppresion of the informational messages\n";
  cerr << "  -h, --help    : prints these informations and exits\n";
  cerr << "  -v, --version : prints the version info and exits\n\n";
  cerr << "  The host defaults to the USA NIST server " << nist
       << '\n';
  cerr << "  (NIST is the National Institute of Standards and "
       << "Technology in Boulder,\n";
  cerr << "  Colorado); a list of public time servers may be found "
       "under the URL\n";
  cerr << "  http://ntp.isc.org/bin/view/Servers/WebHome .\n\n";
}

const char * processOptions(
  int    argc,
  char * argv[]
) {

  // Process the input options; returns a pointer to the host name, or
  // a NULL pointer in case of errors.

  struct option lOpt[] = {
    {"help",    no_argument, 0, 'h'},
    {"quiet",   no_argument, 0, 'q'},
    {"set",     no_argument, 0, 's'},
    {"tcp",     no_argument, 0, 't'},
    {"udp",     no_argument, 0, 'u'},
    {"version", no_argument, 0, 'v'},
    {0,         0,           0,  0 }
  };
  char sOpt[] = "hqstuv";
  char c;

  // "lOpt" defines the long (GNU-style) options; "sOpt" the
  // corresponding short options.

  while ((c = getopt_long(argc, argv, sOpt, lOpt, 0)) != -1) {
    switch (c) {
      case 'h':  usage(argv[0]);        return 0;
      case 'q':  quiet = true;          break;
      case 's':  setTime = true;        break;
      case 't':  st = timeSocket::TCP;  break;
      case 'u':  st = timeSocket::UDP;  break;
      case 'v':  cout << versionID;     return 0;
      case '?':  usage(argv[0]);        return 0;
    }
  }

  if (optind == argc-1) {
    return argv[optind];
  } else if (optind == argc) {
    return nist;
  }

  usage(argv[0]);
  return 0;
}

void resetTime(
  time_t t
) {

  // Resets the current local time to the given value

  struct timeval tv = {t, 0};

  if (settimeofday(&tv, 0) == -1) {
    std::string s("settimeofday: ");
    s += strerror(errno);
    throw std::runtime_error(s);
  }
}

int main(
  int    argc,
  char * argv[]
) {
  const char * host = processOptions(argc, argv);

  if (host) {
    try {
      timeSocket ts(host, st, quiet);
      int32_t    t32 = ts.getTime() - unixTime::unixEpoch;
      unixTime   lTime;
      unixTime   rTime(t32);

      if (!quiet) {
        cout << "Time (local):  " << lTime;
        cout << "\nTime (remote): " << rTime << std::endl;
      }

      if (setTime) {
        if (lTime == rTime) {
          if (!quiet) cout << "The local time is correct\n";
        } else {
          resetTime(rTime.getTime());
          if (!quiet) cout << "The local time has been reset\n";
        }
      }
    } catch (std::exception & e) {
      cerr << "*** EXCEPTION - " << e.what() << '\n';
    }
  }
  return 0;
}
