//     Author: Maurizio Loreti, aka MLO or (HAM) I3NOO
//     Work:   University of Padova - Department of Physics
//             Via F. Marzolo, 8 - 35131 PADOVA - Italy
//     Phone:  +39 (049) 827-7216   FAX: +39 (049) 827-7102
//     EMail:  loreti@pd.infn.it
//     WWW:    http://www.pd.infn.it/~loreti/mlo.html
//
// -------------------------------------------------------------------
// Definition of the classes used by the "synctime" application; see
// the comments in main.cxx for more.
// -------------------------------------------------------------------
// $Id: synctime.hh,v 1.6 2004/02/15 10:07:17 loreti Exp $
// -------------------------------------------------------------------

#ifndef SYNCTIME_HH_
#define SYNCTIME_HH_

#include <iostream>
#include <string>

extern "C" {
  #include <time.h>
  #include <unistd.h>
  #include <sys/socket.h>
  #include <netinet/in.h>
  #include <arpa/inet.h>
}

// A class that encapsulates a socket communication with the
// well-known time port of a trusted time server.
//
// Public interface:
// - the constructor, from the network name of the trusted host, the
//   socket type, and a boolean flag to optionally suppress the
//   informational messages; 
// - an enumerated type, socketType, with the two symbolic constants
//   TCP and UDP, to represent that last value;
// - the destructor;
// - a method, getTime, that queries the time port of the remote host
//   and gets a 32-bits integer containing the number of seconds
//   elapsed after the midnight (hour 0) of January 1, 1900 (the
//   so-called "POSIX time").
//
// - Copy constructor and assignment operator are declared private and
//   never defined, in order to prevent their use: every instanced
//   object owns a resource (the communication socket) released by the
//   destructor, and that is the easy way to prevent unwanted effects
//   destroying temporaries etc.
//
// Private data:
// - _name: the name of the remote time server;
// - _type: the connection type;
// - _quiet: 'true' if the informational messages are suppressed;
// - _socket: the file descriptor of the network connection (or -1 if
//   the connection has not been opened);
// - _host: a structure containing the network data related with the
//   remote host;
// - _timePort: the number of the well-known time port of the remote
//   host (static data member);
// - _timeOut: the number of seconds to wait when trying to get an
//   answer from an UDP connection;
// - a method, getConnection, internally used by getTime to open the
//   socket connected with the remote host.

class timeSocket {
public:
  enum socketType {UDP = 42, TCP};

  timeSocket(const std::string & host, socketType type = TCP,
             bool quiet = false);
  ~timeSocket() { if (_socket >= 0) close(_socket); }
  int32_t getTime();

private:
  std::string        _name;
  socketType         _type;
  bool               _quiet;
  int                _socket;
  struct sockaddr_in _host;

  static const int          _timePort;
  static const unsigned int _timeOut;

  void getConnection();

  timeSocket(const timeSocket &);
  timeSocket & operator = (const timeSocket &);
};

// A method to print on an output stream the value of a variable of
// the type "timeSocket::socketType"

inline std::ostream & operator << (
  std::ostream &           o,
  timeSocket::socketType & type
) {
  o << (type == timeSocket::UDP ? "UDP" : "TCP");
  return o;
}

// A class that encapsulates a time in the standard Unix format:
// i.e. expressed as the number of seconds elapsed since the Unix
// epoch, started at the midnight (hour 0) of January 1, 1970.
//
// Public interface:
// - constructors: from a time_t value (i.e. from a standard Unix
//   time, *not* from a POSIX time), and the default constructor (that
//   uses the current local time);
// - destructor;
// - a method, getTime, that returns the internally stored Unix time;
// - a method, toString, that converts the Unix time in a printable
//   string;
// - an operator used to compare two instances of the class for
//   equality;
// - unixEpoch, the offset between Unix times and POSIX times (static
//   data member).
//
// Private part:
// - _time: the time in standard Unix format;
// - the method setCurrent, that resets the stored time to the current
//   local time;
// - a string, printFormat, that will hold the localized format to be
//   used to print a date/time value with strftime(3).  printFormat is
//   a static variable, and used only if the preprocessor symbol
//   NEED_TZ is defined (see the comments in synctime.cxx)

class unixTime {
private:
  time_t _time;

  void setCurrent() { _time = time(0); }

#ifdef NEED_TZ
  static std::string printFormat;
#endif // NEED_TZ

public:
  static const time_t unixEpoch;

  unixTime() { setCurrent(); }
  unixTime(time_t t) : _time(t) { }
  ~unixTime() { }
  time_t getTime() const { return _time; }
  std::string toString() const;
  bool operator == (const unixTime & rhs) const {
    return _time == rhs._time; }
};

// A method to print on an output stream the value of an instantiation
// of the unixTime class

inline std::ostream & operator << (
  std::ostream &   o,
  const unixTime & t
) {
  o << t.toString();
  return o;
}

#endif // SYNCTIME_HH_
