/*------------------------------------------------------+
 | Author: Maurizio Loreti, aka MLO or (HAM) I3NOO      |
 | Work:   University of Padova - Department of Physics |
 |         Via F. Marzolo, 8 - 35131 PADOVA - Italy     |
 | Phone:  +39 (049) 827-7216   FAX: +39 (049) 827-7102 |
 | EMail:  loreti@pd.infn.it                            |
 | WWW:    http://www.pd.infn.it/~loreti/mlo.html       |
 +------------------------------------------------------+------------

 "synctime" is a program that queries a trusted time server for the
 current time, and compares the returned value with the time on the
 local node.  On request, the time on the local node may be reset.

 synctime has been inspired from "rdata"; however, is a project
 independently developed.

 --------------------------------------------------------------------
 $Id: synctime.c,v 1.11 2006/11/24 10:22:34 loreti Exp $
 -------------------------------------------------------------------*/

#include <errno.h>              /* Standard library headers */
#include <locale.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <sys/types.h>          /* UNIX/POSIX headers */
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/time.h>
#include <signal.h>

#include <langinfo.h>           /* Misc. headers */
#include <getopt.h>

/**
 | TIME_PORT is the well known-port to be queried for time servers;
 | EPOCH is the offset between the Unix epoch and the POSIX time (see
 | later); BUFFER_LENGTH is the dimension of the two internal buffers
 | used for format and printouts; VERSION is the program ID string;
 | NIST the default target node; and UDP_TIMEOUT the timeout value (in
 | seconds) for UDP "recvfrom" calls.
**/

#define TIME_PORT       37
#define EPOCH           2208988800UL
#define BUFFER_LENGTH   64
#define VERSION         "synctime v6 - MLO 15-Feb-2004"
#define NIST            "time-a.timefreq.bldrdoc.gov"
#define UDP_TIMEOUT     10

/**
 | Global variables and local procedure prototypes
**/

int quiet = 0;
int set = 0;
int udp = 0;
static char nist[] = NIST;

int aToAddr(char *, struct sockaddr_in *);
void alarmHandler(int);
int32_t getTime(int, struct sockaddr_in *, char *);
void printTimeT(const char *, const char *, const time_t *);
char *processOptions(int, char * []);
void usage(char *);

#if defined(__sun__)
  int inet_aton(const char *, struct in_addr *);
#endif /* __sun__ */

int main(
  int argc,
  char *argv[]
) {
  struct sockaddr_in hostAddr;
  char format[BUFFER_LENGTH];
  char *host;
  int sock;
  int32_t t32;
  time_t lNow, tNow;
  long int delta;
  const char zone[] = " %Z";

  /**
   | Checks the command line arguments, and stores in "host" a pointer
   | to the trusted time server name
  **/

  host = processOptions(argc, argv);

  /**
   | Converts the trusted host name to an in_addr structure, and
   | stores into that structure the well-known number of the time port
  **/

  if (aToAddr(host, &hostAddr) == 0) {
    fprintf(stderr, "%s: invalid host name\n", host);
    return EXIT_FAILURE;
  }
  hostAddr.sin_family = AF_INET;
  hostAddr.sin_port = htons(TIME_PORT);

  /**
   | Modifies the program locale according to the relevant environment
   | variables (ignoring errors, if any); and gets a localized format
   | string suitable for date/time printout through strftime.  Adds
   | the timezone to the format string, if it is not already there.
  **/

  (void) setlocale(LC_ALL, "");
  strcpy(format, nl_langinfo(D_T_FMT));
  if (strstr(format, zone+1) == 0) strcat(format, zone);

  /**
   | Gets a socket file descriptor from the operating system
  **/

  if (udp) {
    sock = socket(AF_INET, SOCK_DGRAM, 0);
  } else {
    sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
  }

  if (sock < 0) {
    perror("socket");
    return EXIT_FAILURE;
  }

  if (!quiet) {
    fprintf(stdout, "Connecting via %s to %s on port %d ...\n",
           (udp ? "UDP" : "TCP"), inet_ntoa(hostAddr.sin_addr),
           ntohs(hostAddr.sin_port));
  }

  /**
   | Connects to the remote time server, and gets its time
  **/

  t32 = getTime(sock, &hostAddr, host);

  /**
   | Gets and prints the current time on the local host
  **/

  lNow = time(0);
  if (!quiet) printTimeT("Current local time: %s\n", format, &lNow);

  /**
   | Translates the time returned from the trusted server in seconds
   | since the Unix epoch (started at the midnight of January 1,
   | 1970), and computes the difference with the time on the local
   | host.
  **/

  tNow = ntohl(t32) - EPOCH;
  delta = tNow - lNow;

  if (tNow != lNow) { /* if (time should be adjusted) */

    /**
     | The time returned from the trusted time server is different
     | from the time on the local host.  If the -s option has been
     | specified, the time on the local host is synchronized with the
     | trusted host; if not, a warning message is printed.
    **/

    if (set) { /* if (time adjustment has been required) */
      struct timeval t = {0, 0};

      t.tv_sec = tNow;
      if (settimeofday(&t, NULL) != -1) {
        if (!quiet) {
          lNow = time(0);
          fprintf(stdout, "Adjusted by %ld second%s,", delta,
                  (delta == 1 ? "" : "s"));
          printTimeT(" to %s\n", format, &lNow);
        }
      } else {
        perror("settimeofday");
        fprintf(stderr, "Could not adjust time by %ld second%s\n",
                delta, (delta == 1 ? "" : "s"));
      } /* if (time adjustment was successful) */

    } else { /* if (time adjustment has been required) */

      if (!quiet) {
        fprintf(stdout, "(should be adjusted by %ld second%s,",
                delta, (delta == 1 ? "" : "s"));
        printTimeT(" to %s)\n", format, &tNow);
      }
    } /* if (time adjustment has been required) */

  } else {

    if (!quiet) {
      fprintf(stdout, "(is the same time returned from %s)\n",
              host);
    }
  } /* if (time should be adjusted) */

  /**
   | The socket is closed, and the program exits
  **/

  close(sock);
  return EXIT_SUCCESS;
}

void usage(
  char *progName
) {
  printf("Usage: %s [options] [host]\n", progName);
  puts("  Queries the time port of the given host for the "
       "current date and time\n");
  puts("Options:");
  puts("  -q, --quiet   : suppresion of the informational messages");
  puts("  -s, --set     : changes the current time on the local "
       "host");
  puts("  -t, --tcp     : connects to the remote host via a TCP "
       "socket (default)");
  puts("  -u, --udp     : connects to the remote host via an UDP "
       "socket");
  puts("  -h, --help    : prints these informations and exits");
  puts("  -v, --version : prints the version info and exits\n");
  printf("  The host defaults to the USA NIST server %s\n", nist);
  puts("  (NIST is the National Institute of Standards "
       "and Technology in Boulder,");
  puts("  Colorado); a list of public time servers may be found "
       "under the URL");
  puts("  http://ntp.isc.org/bin/view/Servers/WebHome .\n");
}

int aToAddr(
  char *address,
  struct sockaddr_in *pSaddr
) {
  struct hostent *host;

  /**
   | Converts the argument (the trusted host address) to an in_addr
   | structure; the string is first tried as a numeric identifier
   | using inet_aton, then as an alphanumeric identifier using
   | gethostbyname.  Returns 1 if the conversion is successful, 0 if
   | not.
  **/

  if (inet_aton(address, &(pSaddr->sin_addr)) != 0) return 1;

  if ((host = gethostbyname(address)) != NULL) {
    if (host->h_addrtype != AF_INET) {
      fprintf(stderr, "%s: not an AF_INET address\n", address);
      return 0;
    }
    memcpy(&(pSaddr->sin_addr), host->h_addr_list[0], host->h_length);
    return 1;
  }
  return 0;
}

int32_t getTime(
  int sock,
  struct sockaddr_in *pH,
  char *hostName
) {
  int32_t t32;

  if (udp) {
    socklen_t bytes_read = sizeof(*pH);

    /**
     | UDP protocol.  Establishes a signal handler for SIGALRM; sends
     | an empty message to the time port of the remote host, and reads
     | back 32 bits of data containing the time in seconds elapsed
     | after the midnight of January 1, 1900.  The SIGALRM is user to
     | set a timeout.
    **/

    if (signal(SIGALRM, alarmHandler) == SIG_ERR) {
      perror("signal");
      close(sock);
      exit(EXIT_FAILURE);
    }

    if (sendto(sock, NULL, 0, 0, (const struct sockaddr *) pH,
	       sizeof(*pH))) {
      perror("sendto");
      close(sock);
      exit(EXIT_FAILURE);
    }

    alarm(UDP_TIMEOUT);
    if (recvfrom(sock, &t32, sizeof(t32), 0, (struct sockaddr *) pH,
		 &bytes_read) != sizeof(t32)) {
      if (bytes_read > 0) {
        perror("recvfrom");
      } else {
        fprintf(stderr, "EOF received from %s\n", hostName);
      }
      close(sock);
      exit(EXIT_FAILURE);
    }
    alarm(0);

  } else {
    size_t bytes_read = 0;
    int this_read;
    char *buf = (char *) &t32;

    /**
     | TCP protocol; opens a connection with the time port of the
     | remote host, and reads 32 bits of data (with the same meaning)
    **/

    if (connect(sock, (const struct sockaddr *) pH, sizeof(*pH))) {
      perror("connect");
      close(sock);
      exit(EXIT_FAILURE);
    }

    while (bytes_read < sizeof(t32)) {
      do {
        this_read = read(sock, buf, sizeof(t32) - bytes_read);
      } while (this_read < 0  &&  errno == EINTR);

      if (this_read < 0) {
        perror("read");
        close(sock);
        exit(EXIT_FAILURE);
      } else if (this_read == 0) {
        break;
      } else {
        bytes_read += this_read;
        buf += this_read;
      }
    }
    if (bytes_read != sizeof(t32)) {
      fputs("read error\n", stderr);
      close(sock);
      exit(EXIT_FAILURE);
    }
  }
  return t32;
}

void printTimeT(
  const char *headFmt,
  const char *format,
  const time_t *pT
) {
  char buffer[BUFFER_LENGTH];
  struct tm *ptm;

  ptm = localtime(pT);
  strftime(buffer, BUFFER_LENGTH, format, ptm);
  fprintf(stdout, headFmt, buffer);
}

char *processOptions(
  int argc,
  char *argv[]
) {
  struct option lOpt[] = {
    {"help", no_argument, 0, 'h'},
    {"quiet", no_argument, &quiet, 1},
    {"set", no_argument, &set, 1},
    {"tcp", no_argument, &udp, 0},
    {"udp", no_argument, &udp, 1},
    {"version", no_argument, 0, 'v'},
    {0, 0, 0, 0}
  };
  char sOpt[] = "hqstuv";
  char c;

  /**
   | "lOpt" has the long (GNU-style) options; "sOpt" the corresponding
   | short options.  The returned value is a pointer to the host name.
  **/

  while ((c = getopt_long(argc, argv, sOpt, lOpt, NULL)) != -1) {
    switch (c) {
      case 'h':  usage(argv[0]);  exit(EXIT_SUCCESS);
      case 'q':  quiet = 1;  break;
      case 's':  set = 1;  break;
      case 't':  udp = 0;  break;
      case 'u':  udp = 1;  break;
      case 'v':  puts(VERSION);  exit(EXIT_SUCCESS);
      case '?':  usage(argv[0]);  exit(EXIT_FAILURE);
    }
  }

  if (optind == argc-1) {
    return argv[optind];
  } else if (optind == argc) {
    return nist;
  } else {
    usage(argv[0]);
    exit(EXIT_FAILURE);
  }
}

void alarmHandler(
  int sigNum
) {
  /**
   | Called if, in an UDP connection, too much time has elapsed
  **/

  if (sigNum == SIGALRM) {
    fputs("recvfrom: connection timed out\n", stderr);
    exit(EXIT_FAILURE);
  }
}

#if defined(__sun__)
int inet_aton(
  const char     *pc,
  struct in_addr *pin
) {
  in_addr_t addr = inet_addr(pc);
  if (addr == (in_addr_t) -1) {
    return 0;
  } else {
    pin->s_addr = addr;
    return 1;
  }
}
#endif /* __sun__ */
